/*
 $Log: print.C,v $
 * Revision 1.12  1993/11/21  13:33:27  root
 * *** empty log message ***
 *
 * Revision 1.11  1993/11/21  02:14:49  root
 * Option -wc; Option -79x66
 *
 * Revision 1.10  1993/11/21  00:59:49  root
 * US Letter and US Legal paper sizes.
 * Copyright notice.
 * Printing to stdout.
 *
 * Revision 1.9  1993/08/29  19:52:27  root
 * reset printer before and after use
 * fixed getpage (suppressing partially filled logical pages)
 *
 * Revision 1.7  1993/04/05  21:43:36  root
 * changed default page size from 80x66 to 79x66 so that
 * two logical pages fit on one physical page when printing small
 * fixed a bug in print_file()
 *
 * Revision 1.6  1993/04/05  19:10:48  root
 * logical pages
 * correct handling of backspaces: we now can print manpages
 *
 * Revision 1.5  1993/04/05  13:01:35  root
 * rework of squeeze_line(), getpage()
 * changed page_t
 *
 * Revision 1.4  1993/04/03  18:29:04  root
 * changed from German to English
 * improved the white-space-compression
 * more comments
 *
 * Revision 1.3  1993/02/02  02:00:55  root
 * Zeilen- und Spaltenberechnung neu.
 * Verarbeitung von Dateinamen, die mit dem '-' anfangen.
 *
 * Revision 1.2  1993/02/01  02:54:24  root
 * Mehrspaltendruck.
 *
 * Revision 1.1  1993/01/29  00:07:28  root
 * Initial revision
 *
 */

char rcsid[] = "$Id: print.C,v 1.12 1993/11/21 13:33:27 root Exp root $";
char prog_version[] = "1.0";

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

enum {
  opt_lndscp	= 1,
  opt_lq	= 2,
  opt_italic	= 4,
  opt_narrow	= 8,
  opt_lowhgt	= 16,
  opt_phypage	= 32,
  opt_showsize  = 64,

  mode_pipethrough = 1 << 31,

  font_courier	= 1,
  font_cgtimes	= 2,
  font_lettgoth	= 3,

  size_dina4    = 1,
  size_usletter = 2,
  size_uslegal  = 3,

  system_tab	= 8,
  printer_tab	= 8,
  backspc	= 8
};

typedef char * char_p_t;
// This typedef is needed, because `xxx = new (char *)[no_of_lines]' results
// in a warning, and we don't like these.

typedef struct page_t_struct {
  //=============================
  // For each line:
  char_p_t
    *data;	// Space for the text.
  int
    *startcol,	// Physical character column no., the line
		// starts at on the page.
    *len,	// Length of text line with EXPANDED TAB characters.
    		//                          ~~~~~~~~~~~~ see squeeze_line().
  //=============================
  // For the whole page:
    cpi,	// Characters per inch.
    lpi,	// Lines per inch.
    // How many lines and columns fit onto a sheet of paper?
    phy_pgwdt,	// Page width (no. of characters).
    phy_pghgt,	// Page height (no. of lines).
    // Let's divide the page into logical blocks:
    log_lines,	// Logical amount of lines that fit on a page:
	      	// = log_x_pgno * phy_pghgt
    log_pghgt,	// Logical page height. A logical page is guaranteed not to
    		// be broken on text column or physical page boundaries.
    log_y_pgno,	// How many logical pages fit onto a sheet of paper in height?
    log_x_pgno,	// ... and in width?
    log_pgwdt,	// Width of a logical page.
    usedlin,	// No. of lines that contain valid data in actual page.
    linefull,	// Used character columns of printer page. This can be less
	      	// than the whole character width, when using multiple text
	      	// column output.
    datasize;	// The extent of the page buffer.
  char
    *database;	// Start address of character data space to store the
	      	// text lines in.
  //=============================
  // Count all pages:
  int
    log_rd_pg,	// No. of pages read in (logical pages)
    phy_rd_pg,	//			(physical pages)
    log_wr_pg,	// No. of printed pages (logical pages)
    phy_wr_pg;	//			(physical pages)

  //=============================
} page_t;

typedef struct pagelist_t_struct {
  int
    startpage,
    endpage;
  struct pagelist_t_struct
    *next;
} pagelist_t;

// Initialize page structure to be used by get_page() and put_page().
void setup_page(page_t *, unsigned long, unsigned long, unsigned long, int, int);

// Delete dynamically allocated memory of page structure.
void delete_page(page_t *);

// Clear the page without printing it.
void clear_page(page_t *pg);

void print_file(page_t *, pagelist_t *, char *, FILE *, long);

int page_in_pagelist(int, pagelist_t *);

void putpage(page_t *pg, pagelist_t *pgl, FILE *stream);

void setup_pagelist(char *, pagelist_t **);

void delete_pagelist(pagelist_t *);

void issue_ff(FILE *printer);

void setup_dj500(unsigned long, unsigned long, unsigned long,
		 page_t, FILE *);

void reset_dj500(FILE *);

#ifndef DEFAULT_PRT_NAME
#define DEFAULT_PRT_NAME "/dev/lp1"
#endif

#ifndef DEFAULT_FONT
#define DEFAULT_FONT font_courier
#endif

#ifndef DEFAULT_SIZE
#define DEFAULT_SIZE size_dina4
#endif

char default_prt_name[] = DEFAULT_PRT_NAME;
int  default_font       = DEFAULT_FONT;
int  default_size       = DEFAULT_SIZE;

char
  *progname;

int
  file_ok	 = 0,
  file_erroneous = 0;

FILE
  *printer = NULL;

void exit_500(int i)
{
  reset_dj500(printer);
  exit(i);
}

int main(int argno, char *arglist[])
{
  int
    log_x_pgno = 0,
    log_pgwdt = 0,
    i,
    ch;
  unsigned long
    pagesize,
    font,
    flags;
  char
    null_string = 0,
    *pagelist_string,
    *printer_name;
  pagelist_t
    *pagelist;
  page_t
    page = {0, };
  
  progname = arglist[0];

  pagelist_string = &null_string;
  printer_name	  = default_prt_name;
  font	      	  = default_font;
  pagesize        = default_size;
  flags	      	  = 0;

  for(i = 1; i < argno; i += 1) {
    if(strcmp(arglist[i], "-wc") == 0) {
      printf("\n\
    djscript - A text formatter for the HP DeskJet500 (tm) printer.\n\
\n\
    Version %s; Copyright (C) 1993 Joerg Bullmann\n\
\n\
    This program is free software; you can redistribute it and/or modify\n\
    it under the terms of the GNU General Public License as published by\n\
    the Free Software Foundation; either version 2 of the License, or\n\
    (at your option) any later version.\n\
\n\
    This program is distributed in the hope that it will be useful,\n\
    but WITHOUT ANY WARRANTY; without even the implied warranty of\n\
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n\
    GNU General Public License for more details.\n\
\n\
    You should have received a copy of the GNU General Public License\n\
    along with this program; if not, write to the Free Software\n\
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.\n\n", prog_version);
      exit_500(0);
    }
    else if(strncmp(arglist[i], "-P", 2) == 0)
      printer_name = arglist[i] + 2;
    else if(strncmp(arglist[i], "-F", 2) == 0) {
      if(strcmp(arglist[i] + 2, "courier") == 0)
	font = font_courier;
      else if(strcmp(arglist[i] + 2, "cgtimes") == 0)
	font = font_cgtimes;
      else if(strcmp(arglist[i] + 2, "lettgoth") == 0)
	font = font_lettgoth;
      else {
	fprintf(stderr, "%s: unknown font `%s'\n",
		progname, arglist[i] + 2);
	exit_500(1);
      }
    }
    else if(strncmp(arglist[i], "-S", 2) == 0) {
      if(strcmp(arglist[i] + 2, "dina4") == 0)
	pagesize = size_dina4;
      else if(strcmp(arglist[i] + 2, "uslegal") == 0)
	pagesize = size_uslegal;
      else if(strcmp(arglist[i] + 2, "usletter") == 0)
	pagesize = size_usletter;
      else {
	fprintf(stderr, "%s: unknown page size `%s'\n",
		progname, arglist[i] + 2);
	exit_500(1);
      }
    }
    else if(strncmp(arglist[i], "-cw", 3) == 0) {
      if(sscanf(arglist[i] + 3, "%d", &log_pgwdt) != 1 ||
	 log_pgwdt < 0) {
	fprintf(stderr, "%s: bad column width `%s'.\n",
		progname, arglist[i] + 3);
	exit_500(1);
      }
    }
    else if(strncmp(arglist[i], "-cn", 3) == 0) {
      if(sscanf(arglist[i] + 3, "%d", &log_x_pgno) != 1 ||
	 log_x_pgno < 0) {
	fprintf(stderr, "%s: bad column number `%s'.\n",
		progname, arglist[i] + 3);
	exit_500(1);
      }
    }
    else if(strncmp(arglist[i], "-lph", 4) == 0) {
      if(sscanf(arglist[i] + 4, "%d", &(page.log_pghgt)) != 1 ||
	 log_x_pgno < 0) {
	fprintf(stderr, "%s: bad logical page height `%s'.\n",
		progname, arglist[i] + 3);
	exit_500(1);
      }
    }
    else if(strcmp(arglist[i], "-7966") == 0) {
      log_x_pgno     = 0;
      log_pgwdt	     = 79;
      page.log_pghgt = 66;
    }
    else if(strncmp(arglist[i], "-pr", 3) == 0)
      pagelist_string = arglist[i] + 3;
    else if(strcmp(arglist[i], "-s") == 0)
      flags |= opt_showsize;
    else if(strcmp(arglist[i], "-pp") == 0)
      flags |= opt_phypage;
    else if(strcmp(arglist[i], "-ls") == 0)
      flags |= opt_lndscp;
    else if(strcmp(arglist[i], "-itl") == 0)
      flags |= opt_italic;
    else if(strcmp(arglist[i], "-lq") == 0)
      flags |= opt_lq;
    else if(strcmp(arglist[i], "-low") == 0)
      flags |= opt_lowhgt;
    else if(strcmp(arglist[i], "-nrw") == 0)
      flags |= opt_narrow;
    else if(strcmp(arglist[i], "-sml") == 0)
      flags |= (opt_narrow | opt_lowhgt);
    else if(strcmp(arglist[i], "-h") == 0) {
      printf("Usage:\n");
      printf("  %s -h\n", progname);
      printf("  %s [Options] [<fil> ... ]\n", progname);
      printf("Options:\n");
      printf("  -F<font>  : Select font: `courier', `cgtimes', `lettgoth'\n"
	     "              (default: `%s')\n",
	     default_font == font_courier ? "courier" :
	       (default_font == font_lettgoth ? "lettgoth" :
		                                    "cgtimes"));
      printf("  -P<prfil> : Print to file <prfil>.\n"
	     "              (default: `%s', -P- prints to standard output)\n",
	     default_prt_name);
      printf("  -S<size>  : Select paper size: `dina4', `uslegal', `usletter'\n"
	     "              (default: `%s')\n",
	     default_size == size_dina4 ? "dina4" :
	       (default_size == size_uslegal ? "uslegal" :
		                                    "usletter"));
      printf("  -lq       : Letter Quality instead of Draft.\n");
      printf("  -itl      : Italics font.\n");
      printf("  -nrw      : Narrow printing. `horizontal double density'\n"
	     "              (ignored with font `cgtimes')\n");
      printf("  -low      : Low font. `vertical double density'\n");
      printf("  -sml      : Small font: narrow and low.\n");
      printf("  -ls       : Landscape orientation instead of portrait.\n"
	     "              (works only with font `courier normal')\n");
      printf("  -lph<n>   : Set the logical page height to <n> lines.\n");
      printf("  -pp       : Start each file on a new physical page.\n");
      printf("  -cn<n>    : Produce multi column output: <n> columns.\n");
      printf("  -cw<n>    : Produce multi column output: columns of <n>\n"
	     "              characters width each. Logical page width.\n");
      printf("  -7966     : Logical pages of 66 lines and 79 columns each.\n");
      printf("  -pr<pgs>  : Print pages in page range specification <pgs> only.\n");
      printf("              <pgs> is a comma separated list of expressions of the\n");
      printf("              form i, -i, i-j, or i- where i and j denote logical\n");
      printf("              page numbers.\n");
      printf("  -s        : Shows the page geometry; does not print.\n");
      printf("  -wc       : Shows nonwarranty and copying info.\n");
      printf("  -h        : Shows this help info.\n");
      printf("\n");
      exit_500(0);
    }
    else if(strcmp(arglist[i], "-") == 0)
      i += 1;
    else if(arglist[i][0] == '-')
      fprintf(stderr, "%s: option `%s' ignored.\n",
	      progname, arglist[i]);
  }

  if(font == font_cgtimes)
    flags |= mode_pipethrough;

  if((flags & opt_lndscp) &&
     ((flags & opt_italic) || font != font_courier)) {
    fprintf(stderr,
	    "%s: in landscape orientation, font `courier normal' only.\n",
	    progname);
    exit_500(1);
  }

  if(strcmp(printer_name, "-") != 0) {
    if((printer = fopen(printer_name, "w")) == NULL) {
      fprintf(stderr, "%s: couldn't open file `%s'.\n",
	      progname, printer_name);
      exit_500(1);
    }
    fprintf(stderr, "%s: printing to file `%s'.",
	    progname, printer_name);
  }
  else {
    printer = stdout;
    fprintf(stderr, "%s: printing to standard output.", progname);    
  }

  setup_page(&page, font, pagesize, flags, log_x_pgno, log_pgwdt);

  setup_pagelist(pagelist_string, &pagelist);

  setup_dj500(flags, font, pagesize, page, printer);

  for(i = 1; i < argno; i += 1)
    if(arglist[i][0] != '-')
      print_file(&page, pagelist, arglist[i], printer, flags);
    else if(strcmp(arglist[i], "-") == 0 && ++i < argno)
      print_file(&page, pagelist, arglist[i], printer, flags);

  putpage(&page, pagelist, printer);

  if(file_ok + file_erroneous == 0)
    print_file(&page, pagelist, "", printer, flags);
  else {
    if(file_ok > 0)
      fprintf(stderr,
	      "\n%s: %d File(s) printed.",
	      progname, file_ok);
    if(file_erroneous > 0)
      fprintf(stderr,
	      "\n%s: %d File(s) NOT printed.",
	      progname, file_erroneous);
  }

  fprintf(stderr, "\n");

  delete_page(&page);
  delete_pagelist(pagelist);

  reset_dj500(printer);
  fclose(printer);
}

void issue_ff(FILE *printer)
{
  fprintf(printer, "%c%c%c%c%c",
	  0x1b, 0x26, 0x6c, 0x30, 0x48); // page out!
}

void reset_dj500(FILE *printer)
{
  fprintf(printer, "%c%c", 0x1b, 0x45); // Reset the printer!
}

//================================================================
// Initialize the Deskjet 500 and put it into the desired mode.
// The mode is determined by the flags, font, and page variables.
//================================================================
void setup_dj500(unsigned long flags,
		 unsigned long font,
		 unsigned long pagesize,
		 page_t page,
		 FILE *printer)
{
  // HP Deskjet 500 PCL commands

  reset_dj500(printer);

  fprintf(printer, "%c%c%c%c%c",
	  0x1b, 0x26, 0x6b, 0x32, 0x47); // CR => CR + LF

  // Wrap long lines.
  fprintf(printer, "%c%c%c%c%c", 0x1b, 0x26, 0x73, 0x30, 0x43);

  //=============================================================
  // The commands no. 1 to 9 must be issued in this order.
  //=============================================================

  // Paper size
  switch(pagesize) {
  case size_dina4:
    fprintf(printer, "%c%c%c%c%c%c", 0x1b, 0x26, 0x6c, 0x36, 0x26, 0x41);
    break;
  case size_uslegal:
    fprintf(printer, "%c%c%c%c%c", 0x1b, 0x26, 0x6c, 0x33, 0x41);
    break;
  case size_usletter:
    fprintf(printer, "%c%c%c%c%c", 0x1b, 0x26, 0x6c, 0x32, 0x41);
    break;
  default:
    fprintf(stderr, "%s: bad paper size.\n", progname);
    exit_500(1);
  }
  // Perforation skip
  fprintf(printer, "%c%c%c%c%c", 0x1b, 0x26, 0x6c, 0x30, 0x4c);

  // 1. Page orientation
  fprintf(printer, "%c%c%c%c%c",	   // landscape : portrait
	  0x1b, 0x26, 0x6c, (flags & opt_lndscp) ? 0x31 : 0x30, 0x4f);

  // 2. Character set
  fprintf(printer, "%c%c%c%c", 0x1b, 0x28, 0x30, 0x55); // ANSI ASCII

  // 3. Proportional spacing
  fprintf(printer, "%c%c%c%c%c",		   // yes : no
	  0x1b, 0x28, 0x73, (font == font_cgtimes) ? 0x31 : 0x30, 0x50);

  // 4. Horizontal density (cpi)
  if(font != font_cgtimes)
    fprintf(printer, "%c%c%c%d%c", 0x1b, 0x28, 0x73, page.cpi, 0x48);

  // 5. Height of characters
  fprintf(printer, "%c%c%c%d%c", 0x1b, 0x28, 0x73, 18 - page.lpi, 0x56);

  // 6. Typeface
  fprintf(printer, "%c%c%c%c%c",	      // italic : normal
	  0x1b, 0x28, 0x73, (flags & opt_italic) ? 0x31 : 0x30, 0x53);

  // 7. Line thickness normal
  fprintf(printer, "%c%c%c%c%c", 0x1b, 0x28, 0x73, 0x30, 0x42);
  
  // 8. Font
  switch (font) {
    case font_lettgoth:
      fprintf(printer, "%c%c%c%c%c", 0x1b, 0x28, 0x73, 0x36, 0x54);
      break;
    case font_cgtimes:
      fprintf(printer, "%c%c%c%c%c%c%c%c",
	      0x1b, 0x28, 0x73, 0x34, 0x31, 0x30, 0x31, 0x54);
      break;
    default: // font_courier
      fprintf(printer, "%c%c%c%c%c", 0x1b, 0x28, 0x73, 0x33, 0x54);
  }

  // 9. Quality
  fprintf(printer, "%c%c%c%c%c",	    // nice : fast
	  0x1b, 0x28, 0x73, (flags & opt_lq) ? 0x32 : 0x31, 0x51);

  // Vertical spacing (lpi)
  fprintf(printer, "%c%c%c%d%c", 0x1b, 0x26, 0x6c, page.lpi, 0x44);

  // Text Scale Mode
  fprintf(printer, "%c%c%c%c%c", 0x1b, 0x26, 0x6b, 0x36 , 0x57);

}

void setup_pagelist(char *str, pagelist_t **lst)
{
int
  i,
  start = 0, end = 0,
  *x;;
pagelist_t
  *node;

  if(*str == 0) {
    *lst = 0;
    return;
  }

  x = &start;
  for(i = 0; 1; i++) {
    switch(str[i]) {
      case '-':
        if(start == 0) start = 1;
        x = &end;
        break;
      case '0': case '1': case '2': case '3': case '4':
      case '5': case '6': case '7': case '8': case '9':
        *x = *x * 10 + str[i] - '0';
        break;
      case ',':
      case 0:
        if(x == &start) end = start;
        if(start != 0 && (start <= end || end == 0)) {
	  if((*lst = new pagelist_t) == 0) {
	    fprintf(stderr, "%s: not enough memory for pagelist.\n", progname);
	    exit_500(1);
	  }
	  //printf("%d - %d\n", start, end);
	  (*lst)->startpage = start;
	  (*lst)->endpage = end;
	  while(str[i] == ',') i++;
	  setup_pagelist(str + i, &((*lst)->next));
	  return;
	}
	else {
	  fprintf(stderr, "%s: bad pagelist specification `%s'.\n",
		  progname, str);
	  exit_500(1);
	}
        break;
      default:
        fprintf(stderr, "%s: bad pagelist specification `%s'.\n",
		progname, str);
        exit_500(1);
    }
  } 
}

int page_in_pagelist(int pgno, pagelist_t *pgl)
{
  if(pgl == 0) {
    fprintf(stderr, "%d ", pgno);
    return 1;
  }

  while(pgl != 0) {
    if(pgno >= pgl->startpage && (pgno <= pgl->endpage || pgl->endpage == 0)) {
      fprintf(stderr, "%d ", pgno);
      return 1;
    }
    pgl = pgl->next;
  }

  return 0;
}

void delete_pagelist(pagelist_t *pgl)
{
pagelist_t
  *x;

  while(pgl != 0) {
    x = pgl;
    pgl = pgl->next;
    delete x;
  }
}

//===========================================================
// setup_page determines the geometry of the page and the
// text columns. It also allocates memory so that the page
// structure can be used by getpage() and putpage() lateron.
//===========================================================
void setup_page(page_t *pg,
		unsigned long font,
		unsigned long pagesize,
		unsigned long flags,
		int log_x_pgno, int log_pgwdt)
{
  int i, sheetheight, sheetwidth;

  // Reset the page counters
  pg->log_rd_pg = 0;
  pg->log_wr_pg = 0;
  pg->phy_rd_pg = 0;
  pg->phy_wr_pg = 0;

  switch(font) {
    case font_cgtimes:
      pg->cpi = 0;
      break;
    case font_courier:
      pg->cpi = 10;
      break;
    case font_lettgoth:
      pg->cpi = 12;
      break;
    default:
      fprintf(stderr, "%s: bad font in setup_page().\n", progname);
      exit_500(1);
  }

  pg->cpi *= ((flags & opt_narrow) != 0) + 1;
  pg->lpi = ((flags & opt_lowhgt) != 0) ? 12 : 6;

  if(flags & opt_lndscp) {
    // In landscape print mode, the DJ500 has this fixed resolution.
    // It is at least partially contrary to the information given in
    // the German version of the User's Guide. (any clues on that?)
    pg->phy_pgwdt = flags & opt_narrow ? 220 : 110;
    pg->phy_pghgt = flags & opt_lowhgt ? 96 : 48;
  }
  else {
    // For portrait mode I found out these resolutions by trying.
    // Some of the pageheights seem quite "odd" to me: Isn't there
    // more space on a 14 inch sheet of paper than for 72 lines???
    // In lack of "US Legal"-sized sheets, trying is difficult
    // and so I didn't work on that very thoroughly.
    pg->phy_pgwdt = (font == font_courier ? 10 : 12) *
                    (flags & opt_narrow ? 16 : 8);
    switch(pagesize) {
    case size_dina4:
      pg->phy_pghgt = flags & opt_lowhgt ? 139 : 70;
      break;
    case size_usletter:
      pg->phy_pghgt = flags & opt_lowhgt ? 131 : 66;
      break;
    case size_uslegal:
      pg->phy_pghgt = flags & opt_lowhgt ? 143 : 72;
      break;
    default:
      fprintf(stderr, "%s: bad paper size.\n", progname);
      exit_500(1);
    }
  }
  
  sheetheight = pg->phy_pghgt;
  sheetwidth  = pg->phy_pgwdt;

  if(pg->log_pghgt > 0) {
    pg->log_y_pgno  = (pg->phy_pghgt + 1) / (pg->log_pghgt + 1);
    pg->phy_pghgt = pg->log_y_pgno * pg->log_pghgt;
  }
  else {
    pg->log_y_pgno  = 1;
    pg->log_pghgt = pg->phy_pghgt;
  }

  if(log_x_pgno > 0) {
    log_pgwdt = (pg->phy_pgwdt - log_x_pgno + 1) / log_x_pgno;
  }
  else if(log_pgwdt == 0) {
    log_x_pgno = 1;
    log_pgwdt = pg->phy_pgwdt;
  }
  else {
    log_x_pgno = (pg->phy_pgwdt + 1) / (log_pgwdt + 1);
    log_pgwdt = (pg->phy_pgwdt - log_x_pgno + 1) / log_x_pgno;
  }

  if(font == font_cgtimes && log_x_pgno > 1) {
    fprintf(stderr,
	    "%s: Multicolumn output with font `cgtimes' not supported.\n",
	    progname);
    exit_500(1);
  }

  pg->log_lines	= log_x_pgno * pg->phy_pghgt;
  pg->log_x_pgno= log_x_pgno;
  pg->log_pgwdt	= log_pgwdt;
  pg->usedlin	= 0;
  pg->linefull	= log_pgwdt * log_x_pgno + log_x_pgno - 1; // used columns of page

  // Hopefully, this buffersize is big enough:
  pg->datasize = 2 * pg->log_lines * (pg->log_pgwdt + 1);

  if((pg->data     = new char_p_t[pg->log_lines]) == 0 ||
     (pg->database = new char[pg->datasize]) == 0 ||
     (pg->startcol = new int[pg->log_lines]) == 0 ||
     (pg->len	   = new int[pg->log_lines]) == 0) {
    fprintf(stderr,
	    "%s: not enough memory for page buffer.\n",
	    progname);
    exit_500(1);
  }

  pg->data[0] = pg->database;

  if(pg->log_pghgt > pg->phy_pghgt ||
     pg->log_pgwdt > pg->phy_pgwdt) {
    fprintf(stderr,
	    "%s: logical page size exceeds physical page size.\n",
	    progname);
    exit_500(1);
  }

  if(flags & opt_showsize) {
    fprintf(stderr, "\nPaper size: ");
    switch(pagesize) {
    case size_dina4: fprintf(stderr, "DIN A4\n"); break;
    case size_uslegal: fprintf(stderr, "US Legal\n"); break;
    case size_usletter: fprintf(stderr, "US Letter\n"); break;
    }
    fprintf(stderr, "Geometry:       Lines Columns\n");
    fprintf(stderr, "Per inch         %3d    %3d\n", pg->lpi, pg->cpi);
    fprintf(stderr, "On physical page %3d    %3d\n", sheetheight, sheetwidth);
    fprintf(stderr, "On logical page  %3d    %3d\n", pg->log_pghgt, pg->log_pgwdt);
    fprintf(stderr, "Logical pages    %3d    %3d\n", pg->log_y_pgno, pg->log_x_pgno);
    exit_500(0);
  }
}

void delete_page(page_t *pg)
{

  delete pg->database;
  delete pg->data;
  delete pg->startcol;
  delete pg->len;

}

//===================================================================
// squeeze_line preprocesses a text line.
// We want to use as less printer buffer space as possible when
// printing the page lateron.
// We have two main reasons:
// 1) It saves time since we have to wait, if the buffer is full.
// 2) When printing in LANDSCAPE MODE WITH SMALL FONT, a whole
//    page DOESN'T fit into the Deskjet 500's printer buffer (16k).
//    So we have to be accurate here.
// squeeze_line expects the following:
// 1) pg->data[line] is the base address of the string of the current
//    text line. This string needn't be null-terminated, because:
// 2) col contains the length of the text line.
// 3) startcol is the number of leading white spaces. A tab character
//    counts as his adequate space-character-expansion here.
//===================================================================
void squeeze_line(page_t *pg, int line, int col, int startcol)
{
int
  ch,
  poffset,
  leadwhite,
  readcol;

  // Store the number of leading whitespaces.
  // In variable pg->startcol[line] the offset of character
  // columns of the current text column (logical page) is reckoned in.
  pg->startcol[line] =
    (startcol+=((line/pg->log_pghgt)%pg->log_x_pgno)*(pg->log_pgwdt+1));
  //	       ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  // See below in function getpage():
  poffset = 0;
  
  // Cut off trailing white spaces ...
  while(col > 0 && pg->data[line][col - 1] == ' ')
    col--;
  // ... and make the string null-terminated.
  pg->data[line][col] = 0;

  // Replace spaces by appropriate tabs.
  for(col = 0, readcol = 0; pg->data[line][readcol]; readcol += 1, col += 1) {
    ch = pg->data[line][col] = pg->data[line][readcol];
    if(ch == backspc)
      poffset -= 2;
    // The (startcol + readcol + poffset) construct is needed when we are
    // producing multiple column text output.
    if((startcol + readcol + poffset) % printer_tab == printer_tab - 1 &&
       ch == ' ') {
      // We only achieve a saving if we are going backwards here:
      while(col > 0 && pg->data[line][col - 1] == ' ')
	col--;
      pg->data[line][col] = '\t';
    }
  }

  // Determine the length of the line without the leading white spaces.
  pg->len[line]	      = readcol + poffset;
  pg->data[line][col] = 0;

  // The next text line starts in the page buffer at this position:
  if(line + 1 < pg->log_lines)
    pg->data[line + 1] = pg->data[line] + col + 1;

}

//==================================================================
// If the current logical page is empty, do nothing.
// If if is partially filled, fill it up with empty lines.
//==================================================================
void fillpage(page_t *pg)
{
int next_page_start = (pg->usedlin / pg->log_pghgt + 1) * pg->log_pghgt;

  // Is the logical page empty?
  if(next_page_start - pg->log_pghgt == pg->usedlin)
    return;

  // Fill the unused lines of the logical page with `empty lines'.
  while(pg->usedlin < next_page_start)
    squeeze_line(pg, pg->usedlin++, 0, 0);

}

//==================================================================
// Read a page into the page structure *pg.
// If in one getpage() the page isn't filled completely (because of
// an EOF encounter), it can be filled up by immediately following
// getpage()'s.
// If a page is full, additional getpage()'s have NO effect on both
// input stream and page structure.
// getpage() returns the no. of lines in the current page.
// The page buffer is emptied by putpage().
//==================================================================
int getpage(page_t *pg, pagelist_t *pgl, long flags, FILE *stream)
{
int
  poffset,
  leadwhite,
  startcol,
  ch,
  line,
  log_page_startline,
  col;

  // Attach the new data to the end of the stuff that's still
  // in the page buffer. Thus we are able to perform multiple
  // getpage commands with no putpage command between them.
  // If the page is already full, nothing is appended to it.
  if((line = pg->usedlin) == pg->log_lines)
    return line;

  // Start at the beginning of the line.
  col = 0;

  // We don't store the leading white spaces of each line, but
  // their number. So at the start of each line, we're in
  // leading-whitespace-count-mode:
  leadwhite = 1;

  // In variable `startcol' is assumed that the leftmost character column
  // has the number 0. startcol stores the number of leading blanks.
  startcol = 0;

  // If we run into backspace characters, normal column numbers don't
  // correspond to their printing positions on the page anymore.
  // Therefore we introduce the printing offset `poffset', that indicates
  // the difference between the character string column and the print
  // column.
  poffset = 0;

  // This outer loop goes through the logical pages on a physical page.
  do {
    // Determine the first line of the actual logical page.
    log_page_startline = (line / pg->log_pghgt) * pg->log_pghgt;
    // Transfer the logical page from the input stream to the page buffer.
    while(line - log_page_startline < pg->log_pghgt &&
	  (ch = fgetc(stream)) != EOF) {
      if(pg->data[line] + col >= pg->database + pg->datasize) {
	fprintf(stderr, "%s: page buffer overflow.\n", progname);
	exit_500(1);
      }
      else if(startcol + col + poffset >= pg->log_pgwdt ||
              ch == '\f' || ch == '\n') {
	// We interpret a CR character as well as a CTRL-L character
	// to be a line terminating symbol.
	// One line is full.
	squeeze_line(pg, line++, col, startcol);
	// If we have to break the line, we mustn't loose
	// the last character read in.
	if(ch != '\n' && ch != '\f')
	  ungetc(ch, stream);
	col = startcol = poffset = 0;
	leadwhite = 1;
	if(ch == '\f') {
	  // If we've found a formfeed character (CTRL-L)
	  // we skip to the next logical page.
          while(line - log_page_startline < pg->log_pghgt)
	    squeeze_line(pg, line++, col, startcol);
	  // Ignore a CR immediately after a CTRL-L.
	  if((ch = fgetc(stream)) != '\n' && ch != EOF)
	    ungetc(ch, stream);
        }
      }
      else if(leadwhite) {
	if(ch == ' ')
	  startcol++;
	else if(ch == '\t')
	  startcol = (startcol / system_tab + 1) * system_tab;
	else {
	  ungetc(ch, stream);
	  leadwhite = 0;
	}
      }
      else if(ch == '\t') {
	// Expand tabs here. They are `imploded' in squeeze_line().
	// That's necessary because if a line has to be printed in a
	// text column other than the first one (a logical page that
	// is not printed at the left margin of the physical page),
	// a TAB character may have a different meaning 
	// (i.e. space expansion) than expected.
	do {
	  pg->data[line][col++] = ' ';
	} while((startcol + col + poffset) < pg->log_pgwdt &&
		(startcol + col + poffset) % system_tab != 0);
      }
      else if(ch == backspc) {
	if(startcol + col + poffset > 0) {
	  poffset -= 2;
	  pg->data[line][col++] = ch;
	}
	else
	  fprintf(stderr,
		  "%s: warning: trying to backspace over left margin.\n",
		  progname);
      }
      else
	pg->data[line][col++] = ch;

    }

    // If this log. page read finishes because of an EOF encounter, and the last
    // text line read in is neither full nor terminated by a newline character,
    // we have to process it here. If it IS terminated by a newline character,
    // we are dealing with a valid empty line here.
    if(ch == EOF) {
      squeeze_line(pg, line++, col, startcol);
      // Cut off trailing empty lines at the end of the file.
      while(line > 0 && pg->data[line - 1][0] == 0)
	line--;
    }

    if(line > 0) {
      pg->log_rd_pg += 1;
      // Do we have to print it, or is it to be suppressed?
      if(!page_in_pagelist(pg->log_rd_pg, pgl))
        line = ((line - 1) / pg->log_pghgt) * pg->log_pghgt;
    }
  } while(line < pg->log_lines && ch != EOF);

  pg->usedlin = line;
  return line;
}

//==================================================================
// We now have to print out the logical lines of our page in the
// right order so that logical pages and text columns make sense.
//==================================================================
void putpage(page_t *pg, pagelist_t *pgl, FILE *stream)
{
int
  i, j, k,
  log_page_offset,
  line_index,
  thispos,
  previouspos;
char
  null_line = 0,
  *actline;

  if(pg->usedlin == 0) {
    clear_page(pg);
    return;
  }
     
  // Count the pages that are printed.
  pg->phy_wr_pg += 1;
  pg->log_wr_pg += (pg->log_pghgt + pg->usedlin - 1) / pg->log_pghgt;

  // Fill the unused lines of the page with `empty lines'.
  while(pg->usedlin < pg->log_lines)
    pg->data[pg->usedlin++] = &null_line;

  // `log_page_offset' is the number of the logical line that is printed
  // in the upper left corner of the current row of logical pages.

  for(log_page_offset = 0;
      log_page_offset < pg->usedlin;
      log_page_offset += pg->log_pghgt * pg->log_x_pgno) {
    // If this is not the first row of logical pages, separate it
    // from the previous one:
    if(log_page_offset > 0)
      fputc('\n', stream);
    // Traverse the text lines of one row of logical pages on a physical
    // sheet of paper.
    for(i = 0; i < pg->log_pghgt; i++) {
      previouspos = 0;
      line_index  = log_page_offset + i;
      // Process every text column of the current physical line.
      // In one physical line two adjacent logical lines have the
      // logical distance pg->log_pghgt. You can also call this
      // text block distance.
      for(j = 0; j < pg->log_x_pgno; j++, line_index += pg->log_pghgt) {
	actline = pg->data[line_index];
	if(actline[0] != 0) {
	  // If we have an empty logical line, we skip it.
	  thispos = pg->startcol[line_index];
	  // Fill the physical line from previouspos to thispos with spaces!
	  // Don't waste printer buffer space!!!
	  if(thispos - previouspos > 5)
	    fprintf(stream, "%c%c%c%d%c%s",
		    0x1b, 0x26, 0x61, thispos, 0x43, actline);
	  else {
	    for(k = previouspos; k < thispos; k++)
	      fputc(' ', stream);
	    fprintf(stream, "%s", actline);
	  }
	  previouspos = thispos + pg->len[line_index];
	} 
      }
      // Switch to the next physical line.
      if(0 && previouspos < pg->phy_pgwdt) {
        fprintf(stderr, "cr: %d <%s>\n", previouspos, actline);
      }
      if(previouspos < pg->phy_pgwdt)
	fputc('\n', stream);
    }
  }

  // Formfeed.
  // The HP Deskjet 500 is so clever that he doesn't throw away an empty sheet.
  // Thus, if the lines fill the page completely, no additional sheet of paper
  // is released here:
  issue_ff(stream);
  fprintf(stderr, "[%d] ", pg->phy_wr_pg);

  clear_page(pg);
}

void clear_page(page_t *pg)
{
  // Reset the page. Mark it to be empty.
  pg->usedlin = 0;
  pg->data[0] = pg->database;
}

//==================================================================
// Is this page ready to be printed?
// That depends on whether the page ist (at least partially) filled
// and on the flags: flags & opt_phypage means that *pg is to be
// printed even if it is not filled up completely.
// If this flag is not set, the current logical page ist filled
// with blanks.
//==================================================================
int pageready(page_t *pg, long flags)
{
  if(flags & opt_phypage)
    return pg->usedlin > 0;
  else {
    fillpage(pg);
    return pg->usedlin == pg->log_lines;
  }
}

void print_file(page_t *page, pagelist_t *pgl,
		char *filename, FILE *printer,
		long flags)
{
FILE
  *fil;
int
  ch;

  if(filename[0] == 0) {
    fprintf(stderr, "\nstdin - ", progname);
    if(flags & mode_pipethrough)
      while((ch = getc(stdin)) != EOF) 
	putc(ch, printer);
    else {
      while(getpage(page, pgl, flags, stdin) && pageready(page, flags))
	putpage(page, pgl, printer);
    }
    issue_ff(printer);
  }
  else if((fil = fopen(filename, "r")) == NULL) {
    file_erroneous += 1;
    fprintf(stderr, "\n%s: couldn't open file `%s'.",
	    progname, filename);
  }
  else {
    file_ok += 1;
    fprintf(stderr, "\nFile `%s' - ", filename);
    if(flags & mode_pipethrough) {
      while((ch = getc(fil)) != EOF) 
	putc(ch, printer);
      issue_ff(printer);
    }
    else {
      while(getpage(page, pgl, flags, fil) && pageready(page, flags))
	putpage(page, pgl, printer);
    }
    fclose(fil);
  }
}

